/*
 *******************************************************************************
 * touchscreen library : TSlib.c
 *******************************************************************************
 *
 * source code for mikro C compiler V8.2
 * feel free to use this code at your own risks
 *
 * Author : Bruno Gavand, October 2008
 * see more details on http://www.micro-examples.com/
 *
 *******************************************************************************
 */
#include "TSconfig.h"
#include "TSlib.h"

TS_CONFIG tsConfig ;            // touchscreen configuration

/*
 * return 1 if screen is touched, 0 otherwise
 */
unsigned char TS_touched()
        {
        int x ;

        /*
         * activate two opposite plates
         */
        TS_TRIS(0, 0, 0, 1) ;
        TS_PORT(1, 0, 1, 0) ;

        /*
         * read another plate
         */
        x = ADC_Read(TS_ADCYM) ;

        /*
         * deactivate plates
         */
        TS_TRIS(0, 0, 0, 0) ;
        TS_PORT(0, 0, 0, 0) ;

        /*
         * voltage is not null when screen is touched
         */
        if(x > TS_UNTOUCHED)
                {
                return(1) ;
                }

        return(0) ;
        }

/*
 * read row values of X and Y plates
 * return -1 if screen is not touched, 0 otherwise and result is stored
 * in variables pointed to by px and py
 */
signed char   TS_readRaw(int *px, int *py)
        {
        int     x, xx, y, yy ;
        long    l ;

        /*
         * return if screen is not touched
         */
        if(TS_touched() == 0) return(-1) ;

        /*
         * activate X plate
         */
        TS_TRIS(1, 0, 1, 0) ;
        TS_PORT(0, 0, 0, 1) ;
        x = ADC_Read(TS_ADCXP) ;        // read opposite X plate
        TS_PORT(0, 1, 0, 0) ;           // switch X plates
        xx = ADC_Read(TS_ADCXM) ;       // read opposite X plate
        *px = (x - xx) / 2 ;            // average

        /*
         * activate Y plate
         */
        TS_TRIS(0, 1, 0, 1) ;
        TS_PORT(0, 0, 1, 0) ;
        y = ADC_Read(TS_ADCYM) ;        // read opposite Y plate
        TS_PORT(1, 0, 0, 0) ;           // switch Y plates
        yy = ADC_Read(TS_ADCYP) ;       // read opposite Y plate
        *py = (y - yy) / 2 ;            // average

        TS_TRIS(0, 0, 0, 0) ;           // de-activate plates
        TS_PORT(0, 0, 0, 0) ;

        return(0) ;
        }

/*
 * read absolute X and Y coordinates of touched location
 * return -1 if no touch
 * return field number otherwise
 */
signed char   TS_read(int *px, int *py, TS_FIELD *scr)
        {
        int     x, xx, y, yy ;
        long    l ;
        static signed char old = 0 ;

        /*
         * return if screen is not touched
         */
        if(TS_readRaw(px, py) == -1)
                {
                // debounce
                if(old == 1)
                        {
                        Delay_ms(300) ;
                        }
                old = 0 ;
                return(-1) ;
                }

        // debounce
        if(old == 0)
                {
                Delay_ms(20) ;
                old = 1 ;
                return(-1) ;
                }
                
        old = 1 ;
        /*
         * adjust X coordinate
         */
        l = *px - tsConfig.pxmin ;
        l *= SCREEN_WIDTH - 1 ;
        l /= (tsConfig.pxmax - tsConfig.pxmin) ;
        *px = l ;

        /*
         * adjust Y coordinates
         */
        l = *py - tsConfig.pymin ;
        l *= SCREEN_HEIGHT - 1 ;
        l /= (tsConfig.pymax - tsConfig.pymin) ;
        *py = l ;

        /*
         * check for limits
         */
        if(*px < 0) *px = 0 ;
        if(*px >= SCREEN_WIDTH) *px = SCREEN_WIDTH - 1 ;
        if(*py < 0) *py = 0 ;
        if(*py >= SCREEN_HEIGHT) *py = SCREEN_HEIGHT - 1 ;

        /*
         * parse fields if pointer is not null
         */
        if(scr != TS_NULL)
                {
                while(scr->num)
                        {
                        /*
                         * check if touch coordinates are in box
                         */
                        if((*px >= scr->c1) && (*px < scr->c3) && (*py >= scr->c2) && (*py < scr->c4))
                                {
                                /*
                                 * return number
                                 */
                                return(scr->num) ;
                                }

                        scr++ ;         // next field
                        }
                }

        return(0) ;
        }

/*
 * get calibration coordinates
 */
void    TS_calibrate(int *px, int *py)
        {
        /*
         * wait for screen to be touched and get coordinates
         */
        while(TS_readRaw(px, py) == -1) ;

        /*
         * debounce
         */
        Delay_ms(100) ;
        while(TS_touched()) ;
        Delay_ms(100) ;
        }