/*
 *******************************************************************************
 * touchClock : touchscreen library example
 *******************************************************************************
 *
 * This program shows how to use a touchscreen to edit sprites for a clock
 *
 * on power up, program will ask user to touch two points on the screen to calibrate.
 * calibration is stored to Flash memory and will not be asked again, unless calibrate
 * button is pressed on power up.
 * user can edit his own fonts, clock will then display date and time written from user's hand.
 *
 * source code for mikro C compiler V8.2
 * feel free to use this code at your own risks
 *
 * target : PIC18 with 8 Mhz crystal
 *
 * Author : Bruno Gavand, October 2008
 * see more details on http://www.micro-examples.com/
 *
 * Credits to CircuitED for his nice touchscreen :
 * http://www.circuit-ed.com/
 * Thank you Warren !
 *
 *******************************************************************************
 */

#include "timelib.h"            // mikroC time library

#include "TSlib.h"              // touchscreen library definitions
#include "TSconfig.h"           // touchscreen user's configuration

#include "pictures.h"           // pictures bitmap

/*************************
 * REAL TIME CLOCK
 */
/*
 * number of TIMER0 ticks per second
 */
#define T0TPS   (((Clock_Khz() / 4) * 1000) / 256) / 2

TimeStruct      ts ;            // time structure
unsigned int    t0ctr = 0 ;     // timer 0 post-scaler
unsigned long   epoch = 0 ;     // unix epoch

/* ************************************
 * CLOCK SPRITES
 */
#define CONSTBUFSIZE    4736                                    // size of reserved ROM for sprite storage
const char constantBuf[CONSTBUFSIZE] = "" absolute 0x1000 ;     // reserved ROM

#define BLOCK_SIZE      64

/*
 * sprite structure
 */
typedef struct
        {
        unsigned char w, h ;            // width and heigth
        const unsigned char *name ;     // sprite name
        unsigned long addr ;            // sprite address in reserved ROM
        } SPRITE ;

/*
 * sprite widths
 */
#define WIDTH1  8
#define WIDTH2  16
#define WIDTH6  48
#define WIDTH16 128

#define NB_SPRITES       32     // number of sprites

/*
 * sprites table
 */
SPRITE sprite[NB_SPRITES] =
        {
        { WIDTH2, 24, "0" },
        { WIDTH2, 24, "1" },
        { WIDTH2, 24, "2" },
        { WIDTH2, 24, "3" },
        { WIDTH2, 24, "4" },
        { WIDTH2, 24, "5" },
        { WIDTH2, 24, "6" },
        { WIDTH2, 24, "7" },
        { WIDTH2, 24, "8" },
        { WIDTH2, 24, "9" },

        { WIDTH1, 24, "Hour/min separator" },
        { WIDTH1, 24, "Min/sec separator" },

        { WIDTH16, 16, "CLOCK TITLE" },

        { WIDTH6, 24, "Monday" },
        { WIDTH6, 24, "Tuesday" },
        { WIDTH6, 24, "Wednesday" },
        { WIDTH6, 24, "Thursday" },
        { WIDTH6, 24, "Friday" },
        { WIDTH6, 24, "Saturday" },
        { WIDTH6, 24, "Sunday" },

        { WIDTH6, 24, "January" },
        { WIDTH6, 24, "February" },
        { WIDTH6, 24, "March" },
        { WIDTH6, 24, "April" },
        { WIDTH6, 24, "May" },
        { WIDTH6, 24, "June" },
        { WIDTH6, 24, "July" },
        { WIDTH6, 24, "August" },
        { WIDTH6, 24, "September" },
        { WIDTH6, 24, "October" },
        { WIDTH6, 24, "November" },
        { WIDTH6, 24, "December" },
        } ;

unsigned char   currentSprite = 0 ;     // current edited sprite number

/***************************
 * DISPLAY
 */
#define CLOCK_X         8       // time X position, normal operation
#define SET_X           16      // time X position, sprite editing mode
#define DATE_X          24      // date X position
#define CLOCK_Y_TIME    16      // time Y position
#define CLOCK_Y_DATE    40      // date Y position
#define EDITOR_Y        24      // editor box Y position

/***************************
 * TOUCH ZONES DEFINITIONS
 */

/*
 * main screen
 */
TS_FIELD screenDraw[] =
        {
        { 'S', 0, 57, 12, 63 },         // set
        { 'E', 113, 57, 127, 63 },      // edit
        TS_FIELD_END
        } ;

/*
 * date/time setting
 */
TS_FIELD screenSet[] =
        {
        { 'Q', 0, 57, 12, 63 },         // quit
        { 'n', 0, 17,  5, 23 },         // min-
        { 'N', 7, 17, 12, 23 },         // min+
        { 'h', 0, 25,  5, 31 },         // hour-
        { 'H', 7, 25, 12, 31 },         // hour+
        { 'd', 0, 33,  5, 39 },         // day-
        { 'D', 7, 33, 12, 39 },         // day+
        { 'm', 0, 41,  5, 47 },         // month-
        { 'M', 7, 41, 12, 47 },         // month+
        { 'y', 0, 49,  5, 55 },         // year-
        { 'Y', 7, 49, 12, 55 },         // year+
        TS_FIELD_END
        } ;

/*
 * sprite editor
 */
TS_FIELD screenEdit[] =
        {
        { 'X',   0, 57,  12, 63 },      // quit
        { 'C',  14, 57,  26, 63 },      // clear
        { 'I',  28, 57,  38, 63 },      // invert
        { 's',  40, 57,  45, 63 },      // small pen
        { 'L',  47, 57,  53, 63 },      // large pen
        { 'B',  55, 57,  75, 63 },      // black
        { 'W',  77, 57,  95, 63 },      // white
        { 'R',  97, 57, 109, 63 },      // revert
        { '-', 111, 57, 118, 63 },      // previous sprite
        { '+', 120, 57, 127, 63 },      // next sprite
        TS_FIELD_END
        } ;

/**********************
 * OTHER DEFINITIONS
 */
unsigned char setMode = 0 ;     // clock mode: 0 = display date/time, 1 = sprite editor
unsigned int clkX ;             // time X position

/*
 * reset button
 */
#define CALIBRATE_BUTTON        !PORTC.F0

/*
 * interrupt routine handles RTC on timer0 rollover
 */
void    interrupt()
        {
        if(INTCON.TMR0IF)                       // timer 0 overflow ?
                {
                if(setMode == 0)                // if clock is running
                        {
                        t0ctr++ ;               // increment postscaler
                        if(t0ctr == T0TPS)      // one second elapsed ?
                                {
                                epoch++ ;       // increment epoch
                                t0ctr = 0 ;     // reset postscaler
                                }
                        }
                else                            // date/time is being edited
                        {
                        t0ctr = 0 ;             // keep postscaler to 0
                        }
                INTCON.TMR0IF = 0 ;             // reset timer0 rollover flag
                }
        }

/*
 * return constant string length
 * same as strlen
 */
unsigned int    strlenConst(const char *s)
        {
        unsigned int i = 0 ;
        
        while(*s++) i++ ;

        return(i) ;
        }

/*
 * read sprites from flash memory
 * in fact, just calculates sprite adresses and store them to table
 */
void    readSprites()
        {
        unsigned int i ;
        unsigned long addr = (long)constantBuf ;        // reserved rom address

        addr += BLOCK_SIZE ;                            // skip calibration block

        for(i = 0 ; i < NB_SPRITES ; i++)               // for each sprite
                {
                SPRITE *sp ;
                unsigned int x ;

                sp = &sprite[i] ;                       // point to current sprite
                sp->addr = addr ;                       // store address

                /*
                 * calculate next sprite address
                 * with a BLOCK_SIZE alignment
                 */
                for(x = 0 ; x < sp->w * (sp->h / 8) ; x += BLOCK_SIZE)
                        {
                        addr += BLOCK_SIZE ;
                        }
                }
        }

/*
 * write bitmap pointed to by buf to sprite number n
 */
void    writeSprite(unsigned char n, unsigned char *buf)
        {
        unsigned int i ;
        unsigned long addr = (long)constantBuf ;        // reserved ROM address

        addr += BLOCK_SIZE ;                            // skip calibration block

        for(i = 0 ; i < NB_SPRITES ; i++)               // for each sprite
                {
                SPRITE *sp ;
                unsigned int x ;

                sp = &sprite[i] ;                       // point to current sprite

                /*
                 * calculate next sprite address
                 * with a BLOCK_SIZE alignment
                 */
                for(x = 0 ; x < sp->w * (sp->h / 8) ; x += BLOCK_SIZE)
                        {
                        if(i == n)                      // sprite number is reached
                                {
                                Flash_erase_write_64(addr, buf + x) ;   // write block to flash
                                }
                        addr += BLOCK_SIZE ;
                        }
                }
        }

/*
 * display sprite bitmap at address addr of width w and heigth h
 * at position px, py on the GLCD
 */
void    GLCD_spriteWrite(int px, unsigned char py, unsigned long addr, unsigned char w, unsigned char h)
        {
        unsigned char x, y ;

        /*
         * page adjust
         */
        py /= 8 ;
        h /= 8 ;
        
        for(y = py ; y < py + h ; y++)          // for each row
                {
                for(x = px ; x < px + w ; x++)  // for each column
                        {
                        GLCD_set_side(x) ;      // select GLCD side
                        GLCD_set_x(x % 64) ;    // select GLCD column
                        GLCD_set_page(y) ;      // select GLCD page

                        if(addr)                // if addr is not null
                                {
                                GLCD_write_data(*(const unsigned char *)addr++) ;       // write data to GLCD
                                }
                        else
                                {
                                GLCD_write_data(0) ;    // clear GLCD data
                                }
                        }
                }
        }

/*
 * read sprite bitmap of width w and heigth h at position px, py from the GLCD
 * and store it to buffer pointer to by map
 */
void    GLCD_spriteRead(unsigned char px, unsigned char py, unsigned char *map, unsigned char w, unsigned char h)
        {
        unsigned char x, y ;

        /*
         * page adjust
         */
        py /= 8 ;
        h /= 8 ;

        for(y = py ; y < py + h ; y++)          // for each row
                {
                for(x = px ; x < px + w ; x++)  // for each column
                        {
                        GLCD_set_side(x) ;      // select GLCD side
                        GLCD_set_x(x % 64) ;    // select GLCD column
                        GLCD_set_page(y) ;      // select GLCD row
                        GLCD_read_data() ;      // GLCD_read_data() bug workaround : read once for nothing
                        *map++ = GLCD_read_data() ;     // read data and store to buffer
                        }
                }
        }

/*
 * same as GLCD_write_text, but string s is located in ROM
 */
void    GLCD_write_constText(const char *s, unsigned char px, unsigned char py, unsigned char c)
        {
        while(*s)
                {
                GLCD_write_char(*s++, px, py, c) ;
                px += 6 ;       // system font is 5 pixel large + 1 fro spacing
                }
        }

/*
 * edit current sprite
 */
void    spriteEditor()
        {
        int x, y ;
        int     px ;
        signed char ret ;
        unsigned char color = 1 ;
        unsigned char size = 1 ;
        SPRITE *sp ;
        unsigned char   buf[128 * 3] ;

NEW_SPRITE:
        sp = &sprite[currentSprite] ;           // point to current sprite
        px = (SCREEN_WIDTH - sp->w) / 2 ;       // center
        
        GLCD_image(edit_bmp) ;                  // display editor image

        /*
         * cursor above pen size
         */
        switch(size)
                {
                case 0:         // small
                        GLCD_spriteWrite(40, 52, (long)checked, 8, 8) ;
                        break ;
                case 1:         // large
                        GLCD_spriteWrite(47, 52, (long)checked, 8, 8) ;
                        break ;
               }

        /*
         * cursor above color mode
         */
        switch(color)
                {
                case 0:         // black
                        GLCD_spriteWrite(62, 52, (long)checked, 8, 8) ;
                        break ;
                case 1:         // white
                        GLCD_spriteWrite(83, 52, (long)checked, 8, 8) ;
                        break ;
                case 2:         // negate
                        GLCD_spriteWrite(100, 52, (long)checked, 8, 8) ;
                        break ;
                }

        /*
         * display sprite name and draw box
         */
        GLCD_write_constText("Draw sprite for :", (SCREEN_WIDTH - 17 * 6) / 2, 0, 1) ;
        GLCD_write_constText(sp->name, (SCREEN_WIDTH - strlenConst(sp->name) * 6) / 2, 1, 1) ;
        GLCD_rectangle(px - 1, EDITOR_Y - 1, px + sp->w, EDITOR_Y + sp->h, 1) ;

        /*
         * display sprite
         */
        GLCD_spriteWrite(px, EDITOR_Y, sp->addr, sp->w, sp->h) ;

        /*
         * wait for touch panel to be released
         */
        while(TS_touched()) ;

        /*
         * read touchscreen while field is not quit field
         */
        while((ret = TS_read(&x, &y, screenEdit)) != 'X')
                {
                switch(ret)
                        {
                        case -1:                // nothing is touched
                                break ;
                        case 0:                 // no field is touched
                                /*
                                 * inside draw box ?
                                 */
                                if((x >= px) && (x < px + sp->w) && (y >= EDITOR_Y) && (y < EDITOR_Y + sp->h))
                                        {
                                	Glcd_Dot(x, y, color) ;        // yes, draw pixel

                                        if(size)
                                                {
                                                // if big pen, draw sourrounding pixels
                                        	Glcd_Dot(x - 1, y - 1, color) ;
                                        	Glcd_Dot(x    , y - 1, color) ;
                                        	Glcd_Dot(x - 1, y    , color) ;
                                        	Glcd_Dot(x - 1, y + 1, color) ;
                                        	Glcd_Dot(x    , y + 1, color) ;
                                        	}
                                	}
                                break ;

                        case 'C':               // clear
                                memset(buf, 0, sp->w * (sp->h / 8)) ;
                                writeSprite(currentSprite, buf) ;
                                goto NEW_SPRITE ;
                        case 'I':               // invert
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                for(x = 0 ; x < sp->w * (sp->h / 8) ; x++)
                                        {
                                        buf[x] ^= 0xff ;
                                        }
                                writeSprite(currentSprite, buf) ;
                                goto NEW_SPRITE ;

                        case 's':               // select small pen
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                size = 0 ;
                                goto NEW_SPRITE ;
                        case 'L':               // select large pen
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                size = 1 ;
                                goto NEW_SPRITE ;

                        case 'B':               // select black pen
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                color = 0 ;
                                goto NEW_SPRITE ;
                        case 'W':               // select white pen
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                color = 1 ;
                                goto NEW_SPRITE ;
                        case 'R':               // select toggle pen
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                color = 2 ;
                                goto NEW_SPRITE ;

                        case '-':               // previous sprite
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                if(currentSprite > 0)
                                        {
                                        currentSprite-- ;
                                        }
                                else
                                        {
                                        currentSprite = NB_SPRITES - 1 ;
                                        }
                                goto NEW_SPRITE ;
                        case '+':               // next sprite
                                GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
                                writeSprite(currentSprite, buf) ;
                                currentSprite++ ;
                                if(currentSprite == NB_SPRITES)
                                        {
                                        currentSprite = 0 ;
                                        }
                                goto NEW_SPRITE ;
                        }
        	}

        // read and save current sprite
        GLCD_spriteRead(px, EDITOR_Y, buf, sp->w, sp->h) ;
        writeSprite(currentSprite, buf) ;
        }

/*
 * redraw main screen
 */
void    redraw()
        {
        SPRITE  *sp ;

        if(setMode)
                {
                /*
                 * clock setting screen
                 */
                Glcd_Image(setmode_bmp) ;
                clkX = SET_X ;
                }
        else
                {
                /*
                 * clock screen
                 */
                Glcd_Image(background_bmp) ;
                clkX = CLOCK_X ;
                }

        // display clock title
        sp = &sprite[12] ;
        GLCD_spriteWrite((SCREEN_WIDTH - sp->w) / 2, 0, sp->addr, sp->w, sp->h) ;

        // display hours/min. separator
        sp = &sprite[10] ;
        GLCD_spriteWrite(clkX + 2 * WIDTH2, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;

        // display min./sec. separator
        sp = &sprite[11] ;
        GLCD_spriteWrite(clkX + 4 * WIDTH2 + WIDTH1, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;
        }

/*
 * roll up or down char pointed to by v depending on direction dir
 * with min and max values
 */
void    roll(unsigned char *v, signed char dir, unsigned char min, unsigned char max)
        {
        if(dir > 0)
                {
                (*v)++ ;
                if(*v > max)
                        {
                        *v = min ;
                        }
                }
        else
                {
                if(*v > min)
                        {
                        (*v)-- ;
                        }
                else
                        {
                        *v = max ;
                        }
                }
        }

/*
 * main program
 */
void	main()
	{
	int x, y ;
	unsigned char displayMD = 0 ;

        /*
         * PORT configuration
         */
        TRISA = 0 ;
#ifdef P18F4620
	ADCON1 = 0b1111 ;
	CMCON = 7 ;
#endif
#ifdef P18F4685
	ADCON1 = 0b1111 ;
	CMCON = 7 ;
#endif

#ifdef P18F452
	ADCON1 = 0b0111 ;
#endif

	PORTA = 0xff ;

	PORTB = 0 ;
	TRISB = 1 ;
	
	PORTC = 0 ;
	TRISC = 0b00000001 ;

	PORTD = 0 ;
  	TRISD = 0 ;

//	Delay_ms(100) ;

        /*
         * init GLCD
         */
  	Glcd_Init(&PORTB, 2, 3, 4, 5, 7, 6, &PORTD);         // easyPIC3
  	Glcd_Fill(0) ;                                       // clear screen
  	Glcd_Set_Font(FontSystem5x8, 5, 8, 32) ;             // select5x8 fonts

        readSprites() ;                                 // prepare sprites

        /*
         * read touchscreen calibration
         */
        Flash_read_N_bytes((long)constantBuf, (unsigned char *)&tsConfig, sizeof(TS_CONFIG)) ;
        if((tsConfig.pxmin == 0) || CALIBRATE_BUTTON)           // if calibration required
                {
                /*
                 * top left touch
                 */
                GLCD_write_text("TOUCH TOP LEFT", 20, 3, 1) ;
                GLCD_spriteWrite(0, 0, (long)arrowTL, 8, 8) ;
                TS_calibrateTL() ;
                Glcd_Fill(0) ;

                /*
                 * bottom right touch
                 */
                GLCD_write_text("TOUCH BOTTOM RIGHT", 10, 3, 1) ;
                GLCD_spriteWrite(SCREEN_WIDTH - 8, SCREEN_HEIGHT - 8, (long)arrowBR, 8, 8) ;
                TS_calibrateBR() ;

                /*
                 * save to flash memory
                 */
                Flash_erase_write_64((long)constantBuf, (unsigned char *)&tsConfig) ;
                }

        /*
         * draw screen
         */
        redraw() ;

        /*
         * prepare timer 0 and interrupt for real-time clock
         */
        T0CON = 0b11000000 ;    // TMR0ON T08BIT T0CS T0SE PSA T0PS2 T0PS1 T0PS0
        INTCON.TMR0IE = 1 ;
        INTCON.GIE = 1 ;

        for(;;)                                         // forever
                {
                SPRITE *sp ;
                unsigned char p ;

                Time_epochToDate(epoch, &ts) ;          // convert unix epoch to time structure

                /*
                 * display hours
                 */
                sp = &sprite[ts.hh / 10] ;              // tenth
                if(ts.hh / 10)
                        {
                        GLCD_spriteWrite(clkX, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;
                        }
                else
                        {
                        GLCD_spriteWrite(clkX, CLOCK_Y_TIME, 0, sp->w, sp->h) ;
                        }
                sp = &sprite[ts.hh % 10] ;              // units
                GLCD_spriteWrite(clkX + WIDTH2, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;

                /*
                 * display minutes
                 */
                sp = &sprite[ts.mn / 10] ;              // tenth
                GLCD_spriteWrite(clkX + 2 * WIDTH2 + WIDTH1, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;
                sp = &sprite[ts.mn % 10] ;              // units
                GLCD_spriteWrite(clkX + 3 * WIDTH2 + WIDTH1, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;

                /*
                 * display seconds
                 */
                sp = &sprite[ts.ss / 10] ;              // tenth
                GLCD_spriteWrite(clkX + 4 * WIDTH2 + 2 * WIDTH1, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;
                sp = &sprite[ts.ss % 10] ;              // units
                GLCD_spriteWrite(clkX + 5 * WIDTH2 + 2 * WIDTH1, CLOCK_Y_TIME, sp->addr, sp->w, sp->h) ;

                /*
                 * altern week day + month day with month + year
                 * every 4 seconds
                 */
                if(setMode == 0)
                        {
                        if((ts.ss >> 2) % 2)
                                {
                                displayMD = 0 ;
                                }
                        else
                                {
                                displayMD = 1 ;
                                }
                        }
                        
                if(displayMD)
                        {
                        /*
                         * display week day name
                         */
                        p = DATE_X ;
                        sp = &sprite[13 + ts.wd] ;
                        GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;

                        /*
                         * display month day number
                         */
                        p += sp->w ;
                        sp = &sprite[ts.md / 10] ;      // tenth
                        if(ts.md / 10)
                                {
                                GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;
                                }
                        else
                                {
                                GLCD_spriteWrite(p, CLOCK_Y_DATE, 0, sp->w, sp->h) ;
                                }
                        p += sp->w ;
                        sp = &sprite[ts.md % 10] ;      // units
                        GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;
                        }
                else
                        {
                        /*
                         * display month name
                         */
                        p = DATE_X ;
                        sp = &sprite[20 + ts.mo - 1] ;
                        GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;

                        /*
                         * display year
                         */
                        p += sp->w ;
                        sp = &sprite[(ts.yy / 10) % 10] ;       // tenth
                        GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;
                        p += sp->w ;
                        sp = &sprite[ts.yy % 10] ;              // units
                        GLCD_spriteWrite(p, CLOCK_Y_DATE, sp->addr, sp->w, sp->h) ;
                        }

                /*
                 * read touchscreen
                 */
                switch(TS_read(&x, &y, setMode ? screenSet : screenDraw))
                        {
                        case -1:                // no touch
                                break ;

                        case 'E':               // end
                                spriteEditor() ;
                                redraw() ;
                                while(TS_touched()) ;
                                break ;
                        case 'S':               // set mode
                                setMode = 1 ;
                                ts.ss = 0 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                redraw() ;
                                while(TS_touched()) ;
                                break ;

                        case 'Q':               // quit
                                setMode = 0 ;
                                redraw() ;
                                while(TS_touched()) ;
                                break ;
                        case 'Y':               // year+
                                if(ts.yy == 2069)
                                        {
                                        ts.yy = 1970 ;
                                        }
                                else
                                        {
                                        ts.yy++ ;
                                        }
                                displayMD = 0 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'y':               // year-
                                if(ts.yy > 1970)
                                        {
                                        ts.yy-- ;
                                        }
                                else
                                        {
                                        ts.yy = 2069 ;
                                        }
                                displayMD = 0 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'M':               // month+
                                roll(&ts.mo, 1, 1, 12) ;
                                displayMD = 0 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'm':               // month-
                                roll(&ts.mo, -1, 1, 12) ;
                                displayMD = 0 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'D':               // day+
                                roll(&ts.md, 1, 1, 31) ;
                                displayMD = 1 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'd':               // day-
                                roll(&ts.md, -1, 1, 31) ;
                                displayMD = 1 ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'H':               // hour+
                                roll(&ts.hh, 1, 0, 23) ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'h':               // hour-
                                roll(&ts.hh, -1, 0, 23) ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'N':               // minute+
                                roll(&ts.mn, 1, 0, 59) ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                        case 'n':               // minute-
                                roll(&ts.mn, -1, 0, 59) ;
                                epoch = Time_dateToEpoch(&ts) ;
                                while(TS_touched()) ;
                                break ;
                	}
        	}
	}
